#ifndef bqtExprHH
#define bqtExprHH

#include <string>
#include <vector>

#include "tristate"
#include "tokens.hh"
#include "autoptr"

class expr: public ptrable
{
public:
    expr();
    virtual ~expr();
    
    virtual tristate is_num() const = 0;
    virtual tristate is_bool() const = 0;
    virtual tristate is_string() const = 0;
    virtual bool is_const() const = 0;
};
typedef autoptr<expr> exprp;

class expr_unary: public expr
{
    QBtoken token;
    exprp p;
public:
	expr_unary(QBtoken tok, exprp a) : token(tok), p(a) { }

    virtual tristate is_num() const    { return p->is_num(); }
    virtual tristate is_bool() const   { return p->is_bool(); }
    virtual tristate is_string() const { return p->is_string(); }
    virtual bool is_const() const  { return p->is_const(); }
};
class expr_binary: public expr
{
    QBtoken token; 
    exprp p0, p1;
public:
	expr_binary(QBtoken tok, exprp a, exprp b) : token(tok), p0(a), p1(b) { }

    virtual tristate is_num() const    { return maybe; } //FIXME
    virtual tristate is_bool() const   { return maybe; }
    virtual tristate is_string() const { return maybe; }
    virtual bool is_const() const  { return p0->is_const() && p1->is_const(); }
};

class expr_number: public expr
{
    enum { t_int, t_float } type;
    long long int int_val;
    long double float_val;
    // TODO: value here.
public:
    expr_number() : type(t_int), int_val(0), float_val(0) { }
    expr_number(long long int i) : type(t_int), int_val(i) { }
    expr_number(long int i) : type(t_int), int_val(i) { }
    expr_number(long double f) : type(t_float), float_val(f) { }
    expr_number(int i) : type(t_int), int_val(i) { }
    expr_number(double f) : type(t_float), float_val(f) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const  { return true; }
};

class expr_label: public expr
{
    long double label;
public:
    expr_label() : label(0) { }
    expr_label(long double f) : label(f) { }
    expr_label(long long int i) : label(i) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const  { return true; }
    
    long double get_label() const { return label; }
};

class expr_string: public expr
{
    std::string value;
public:
    expr_string() { }
    expr_string(const char* v) : value(v) { }
    expr_string(const std::string& v) : value(v) { }

    virtual tristate is_num() const    { return false; }
    virtual tristate is_bool() const   { return false; }
    virtual tristate is_string() const { return true; }
    virtual bool is_const() const  { return true; }
};

class expr_userfun: public expr
{
    std::string name;
    std::vector<exprp> params;
public:
    expr_userfun(const std::string& n) : name(n) { }    
    
    virtual tristate is_num() const    { return maybe; } // FIXME
    virtual tristate is_bool() const   { return maybe; } // FIXME
    virtual tristate is_string() const { return maybe; } // FIXME
    virtual bool is_const() const  { return false; }

};

class expr_builtin_numeric: public expr
{
    QBtoken token;
    std::vector<exprp> params;
public:
    expr_builtin_numeric(QBtoken tok) : token(tok) { }
    expr_builtin_numeric(QBtoken tok, exprp a) : token(tok)
        { params.push_back(a); }
    expr_builtin_numeric(QBtoken tok, exprp a, exprp b) : token(tok)
        { params.push_back(a); params.push_back(b); }
    expr_builtin_numeric(QBtoken tok, exprp a, exprp b, exprp c) : token(tok)
        { params.push_back(a); params.push_back(b); params.push_back(c); }
    expr_builtin_numeric(QBtoken tok, exprp a, exprp b, exprp c, exprp d) : token(tok)
        { params.push_back(a); params.push_back(b); params.push_back(c); params.push_back(d); }


    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const
    {
        for(unsigned a=0; a<params.size(); ++a)
            if(!params[a]->is_const()) return false; 
        return (token == ABS
             || token == ATN
             || token == CDBL
             || token == CINT
             || token == CLNG
             || token == COS
             || token == CSNG
             || token == EXP
             || token == FIX
             || token == INT
             || token == LOG
             || token == SQR
             || token == TAN
             || token == VAL
               );
    }
};

class expr_builtin_string: public expr
{
    QBtoken token;
    std::vector<exprp> params;
public:
    expr_builtin_string(QBtoken tok) : token(tok) { }
    expr_builtin_string(QBtoken tok, exprp a) : token(tok)
      { params.push_back(a); }
    expr_builtin_string(QBtoken tok, exprp a, exprp b) : token(tok)
      { params.push_back(a); params.push_back(b); }
    expr_builtin_string(QBtoken tok, exprp a, exprp b, exprp c) : token(tok)
      { params.push_back(a); params.push_back(b); params.push_back(c); }
    expr_builtin_string(QBtoken tok, exprp a, exprp b, exprp c, exprp d) : token(tok)
      { params.push_back(a); params.push_back(b);
        params.push_back(c); params.push_back(d); }
    
    virtual tristate is_num() const    { return false; }
    virtual tristate is_bool() const   { return false; }
    virtual tristate is_string() const { return true; }
    virtual bool is_const() const
    {
        for(unsigned a=0; a<params.size(); ++a)
            if(!params[a]->is_const()) return false; 
        
        /* Pure functions: functions that only depend on their parameter;
         * can be calculated at compile-time. */
        return (token == LTRIM_S
             || token == RTRIM_S
             || token == LCASE_S
             || token == UCASE_S
             || token == INSTR
             || token == MID_S
             || token == RIGHT_S
             || token == STRING_S
             || token == CHR_S
             || token == HEX_S
             || token == MKD_S
             || token == MKI_S
             || token == MKL_S
             || token == MKS_S
             || token == OCT_S
             || token == SPACE_S
             || token == STR_S
               );
    }
};

class expr_coordinate: public expr
{
    bool step;
    exprp x, y;
public:
    virtual tristate is_num() const    { return false; }
    virtual tristate is_bool() const   { return false; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const { return x->is_const() && y->is_const(); }
};

/* These can only occur in select-case */
class expr_caseelse: public expr
{
public:
    expr_caseelse(QBtoken) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const { return true; }
};
class expr_caseis: public expr
{
    QBtoken token;
    exprp ref;
public:
    expr_caseis(QBtoken t, exprp r) : token(t), ref(r){ }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const { return ref->is_const(); }
};
class expr_caserange: public expr
{
    exprp min, max;
public:
    expr_caserange(QBtoken, exprp mi, exprp ma) : min(mi), max(ma) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const { return min->is_const() && max->is_const(); }
};
class expr_caseor: public expr
{
    exprp a,b;
public:
    expr_caseor(QBtoken, exprp aa, exprp bb) : a(aa),b(bb) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const { return a->is_const() && b->is_const(); }
};


class expr_var: public expr
{
    std::string name;
    std::vector<exprp> indices;
    autoptr<expr_var> member;            // NULL if none
public:
    expr_var(const std::string& n) : name(n) { }
    
    virtual tristate is_num() const    { return maybe; } // FIXME
    virtual tristate is_bool() const   { return maybe; } // FIXME
    virtual tristate is_string() const { return maybe; } // FIXME
    virtual bool is_const() const { return false; }
    
    void addindex(exprp p) { indices.push_back(p); }
    void addmember(autoptr<expr_var> m) { member = m; }
    
    bool hasmembers() const { return member; }
    bool hasindexes() const { return !indices.empty(); }
};

class expr_lbound: public expr
{
    exprp var;
    exprp num_index;
public:
    expr_lbound(QBtoken, exprp a, exprp b = new expr_number(1))
        : var(a), num_index(b) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }
};

class expr_ubound: public expr
{
    exprp var;
    exprp num_index;
public:
    expr_ubound(QBtoken, exprp a, exprp b = new expr_number(1))
        : var(a), num_index(b) { }

    virtual tristate is_num() const    { return true; }
    virtual tristate is_bool() const   { return true; }
    virtual tristate is_string() const { return false; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }
};


#endif
