#include <deque>
#include <set>
#include <cstdio>

template<typename T>
class Lex
{
    static const long eof_pos = -1;
public:
    class iterator
    {
        Lex* parent;
        long pos;
    public:
        iterator() : parent(0),pos(eof_pos) { }
        iterator(const iterator& ref)
            : parent(ref.parent), pos(ref.pos)
            { if(parent) parent->referrer_list.insert(this); }
        ~iterator()
            { if(parent) { 
              parent->referrer_list.erase(this);
              parent->updatepos();
            } }
        
        const T& operator* () const { return parent->get(pos); }
        iterator& operator++() { inc(); return *this; }
        iterator operator++(int) { iterator tmp(*this); inc(); return tmp; }
        bool operator==(const iterator& b) const { return pos == b.pos; }
        bool operator!=(const iterator& b) const { return !operator==(b); }

        typedef LexLocation LocType;
        const LocType GetLoc() const { return parent->getloc(pos); }

    private:
        friend class Lex;
        iterator(Lex& obj, long p)
            : parent(&obj), pos(p)
            { parent->referrer_list.insert(this); }
        
        void inc()
        {
            ++pos;
            parent->updatepos(); 
            if(parent->eof(pos)) { pos = eof_pos; }
        }
    };
    
    iterator begin() { return iterator(*this, 0); }
    iterator end()   { return iterator(); }
    
    Lex()
        : elements(), referrer_list(), minimum(0), supplementable(true)
    {
    }
    
private:
    bool eof(long pos) const
    {
        return (unsigned)(pos-minimum) >= elements.size() && !supplementable;
    }
    const T& get(long pos) const
    {
#ifdef LEXER_DEBUG
        fprintf(stderr, "Asked token %ld: %s\n",
            (long)(pos),
            GetTokenName(  elements[pos-minimum].second  ).c_str());
#endif
        return elements[pos-minimum].second;
    }
    const LexLocation& getloc(long pos) const
    {
        return elements[pos-minimum].first;
    }
    void updatepos()
    {
        long smallest_pos = 0;
        long biggest_pos = 0;
        bool first = true;
        for(typename std::set<iterator*>::const_iterator
            i = referrer_list.begin();
            i != referrer_list.end();
            ++i)
        {
            if((*i)->pos == eof_pos) continue; // ignore EOFs
            
            if(first) { first=false; smallest_pos = biggest_pos = (*i)->pos; continue; }
            if((*i)->pos < smallest_pos) smallest_pos = (*i)->pos;
            if((*i)->pos >  biggest_pos)  biggest_pos = (*i)->pos;
        }
#ifdef LEXER_DEBUG
        fprintf(stderr, "Updatepos: smallest=%ld, biggest=%ld (%lu referrers)\n",
            smallest_pos, biggest_pos,
            (unsigned long)referrer_list.size());
#endif
        long excess = smallest_pos - minimum;
        if(excess > 0)
        {
#ifdef LEXER_DEBUG
            fprintf(stderr, "discarding %ld excess tokens\n", excess);
#endif
            elements.erase(elements.begin(), elements.begin() + excess);
            minimum = smallest_pos;
        }
        
        while(biggest_pos - minimum >= (long)elements.size())
        {
            LexLocation loc = yyGetLocation();
            T tok = (T)YY_LEXFUNAME();
#ifdef LEXER_DEBUG
            fprintf(stderr, "Token[%ld]: %s\n",
                (long)(elements.size()+minimum),
                GetTokenName(tok).c_str());    
#endif
            if(!tok) { supplementable = false; break; } // EOF
            elements.push_back( std::make_pair(loc, tok) );
        }
    }

private:
    /* copies not supported */
    Lex(const Lex&);
    void operator=(const Lex&);

private:
    std::deque< std::pair<LexLocation,T> > elements;
    std::set<iterator*> referrer_list;
    long minimum;
    bool supplementable;
};
