/* Flex declarations */

%option noyywrap
/*%option debug*/
%option caseless
%option 8bit
%option never-interactive

%x STATE_DEFCMD
%x STATE_COMMENT
%x STATE_INCLUDE
%x STATE_DATA

%{
/* C declarations */
#include <cctype>
#include <cstring>
#include <list>

#include "stringfun.hh"
#include "binding.hh"
#include "lexer.hh"

LexerResult yylval;
static QBtoken prev_return = CRLF;
#define Return return prev_return=(QBtoken)

typedef struct saved_yy_context
{
    LexLocation loc;
    YY_BUFFER_STATE state;
    FILE *fp;
} saved_yy_context;
static saved_yy_context current_context;
static std::list<saved_yy_context> contexts;

#define YY_INPUT(buf,result,max_size) \
  { \
    int c; do c = fgetc(yyin); while(c == '\r'); \
    if(c == EOF) result=YY_NULL; \
    else { \
      if(c == '\n')      { ++current_context.loc.line; current_context.loc.col=1; } \
      else if(c == '\t') { current_context.loc.col=(current_context.loc.col+7)&~7; } \
      else { ++current_context.loc.col; } \
      buf[0]=c; result=1; \
    } \
  }

void yyStartFile(FILE *fp, const char *name)
{
    current_context.loc.line     = 1;
    current_context.loc.col      = 1;
    current_context.loc.filename = name;
    /* .fp and .state are filled at $INCLUDE. */
    yyin = fp;
}
void yyEndFile(void)
{
}
const LexLocation& yyGetLocation()
{
     return current_context.loc;
}

void yyerror(const std::string& msg)
{
    yyerror(msg, current_context.loc);
}
void yyerror(const std::string& msg, const LexLocation& loc)
{
    fprintf(stderr,
        "\n%s:%d:%d: %s\n",
        loc.filename.c_str(),
        loc.line,
        loc.col,
        msg.c_str()
    );
}

static char cur_deftype = '!';
static void DefVarType(int begin, int end)
{
    begin = toupper(begin);
    end   = toupper(end);
    if(begin > end) { int c=begin;begin=end;end=c; } /* ensure correct order */
    
    SetDefVarType( (char)begin, (char)end, cur_deftype);
}

%}
/* Flex declarations */

digit [0-9]
blank [ \t]

%%
	/* Grammar begins: */

	/* Exception: DEFTYPE commands */
DEFINT     BEGIN(STATE_DEFCMD); cur_deftype = '%';
DEFSNG     BEGIN(STATE_DEFCMD); cur_deftype = '!';
DEFDBL     BEGIN(STATE_DEFCMD); cur_deftype = '#';
DEFSTR     BEGIN(STATE_DEFCMD); cur_deftype = '$';
DEFLNG     BEGIN(STATE_DEFCMD); cur_deftype = '&';

<STATE_DEFCMD>[A-Z]-[A-Z] DefVarType(yytext[0], yytext[2]);
<STATE_DEFCMD>[A-Z]       DefVarType(yytext[0], yytext[0]);
<STATE_DEFCMD>[, \t]+ /* Ignore comma and space */
<STATE_DEFCMD>:[ :]*  BEGIN(INITIAL); Return COLONCHAR; /* end of def */
<STATE_DEFCMD>\n+     BEGIN(INITIAL); Return CRLF;      /* end of def */
<STATE_DEFCMD>\'      BEGIN(STATE_COMMENT);


        /* Exception: DATA */
DATA       BEGIN(STATE_DATA); Return DATA;
<STATE_DATA>\"[^"\n]*["\n]  {
    //fprintf(stderr, "Data string (quoted): '%s'\n", yytext);
    yylval.str = (yytext);
    Return STRINGCONST;
  }
<STATE_DATA>[ \t]+    /* Ignore space in DATA */
<STATE_DATA>[^:,\n]+  {
    //fprintf(stderr, "Data string (unquoted): '%s'\n", yytext);
    /* ltrim and rtrim the string - qb does that too. */
    yylval.str = (Trim(yytext));
    Return STRINGCONST;
  }
<STATE_DATA>,      Return *yytext;
<STATE_DATA>:[ :]* BEGIN(INITIAL); Return COLONCHAR; /* end of data */
<STATE_DATA>\n+    BEGIN(INITIAL); Return CRLF;      /* end of data */



	/* Exception: Comments */
REM      { if(prev_return != COLONCHAR && prev_return != CRLF) Return REM;
           /* REM starts a comment only when at statement boundary */
           BEGIN(STATE_COMMENT); }
\'       { BEGIN(STATE_COMMENT); }

<STATE_COMMENT>\$DYNAMIC {
    SetMetaDynamic();
    /* Don't return anything */
  }
<STATE_COMMENT>\$STATIC {
    SetMetaStatic();
    /* Don't return anything */
  }
<STATE_COMMENT>\$INCLUDE:\ *\' BEGIN(STATE_INCLUDE);
<STATE_COMMENT>\n+             BEGIN(INITIAL); Return CRLF;
<STATE_COMMENT>[^$\n]+   /* Ignore longest possible blobs until $ */
<STATE_COMMENT>.         /* Ignore other   */

<STATE_INCLUDE>[^']* {
    fprintf(stderr, "Meta $INCLUDE: '%s'\n", yytext);
    {
      current_context.state = YY_CURRENT_BUFFER;
      current_context.fp    = yyin;
      contexts.push_back(current_context);
      
      yyin = fopen(yytext, "rt");
      if(!yyin) { YY_FATAL_ERROR("unable to open include file"); }
      /*setbuf(yyin, NULL);*/
      
      yyStartFile(yyin, yytext);
      
      yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
      
      BEGIN(INITIAL);
    }
  }
<STATE_INCLUDE>\'        BEGIN(STATE_COMMENT);

	/* Logical operators */
IMP        Return IMP;
EQV        Return EQV;
XOR        Return XOR;
OR         Return  OR;
AND        Return AND;
NOT        Return NOT;

MOD        Return MOD;

	/* If stuff */
IF         Return IF;
THEN       Return THEN;
ELSE       Return ELSE;
ELSEIF     Return ELSEIF;
"END IF"   Return ENDIF;

	/* Switch stuff */
SELECT     Return SELECT;
CASE       Return CASE;
"END SELECT"  Return ENDSELECT;
IS         Return IS;

	/* Loop stuff */
"DO WHILE" Return DOWHILE;
"DO UNTIL" Return DOUNTIL;
"LOOP WHILE" Return LOOPWHILE;
"LOOP UNTIL" Return LOOPUNTIL;
DO         Return DO;
LOOP       Return LOOP;
WHILE      Return DOWHILE;
WEND       Return LOOP;
"EXIT DO"  Return EXITDO;

	/* Sub stuff */
SUB           Return SUB;
"END SUB"     Return ENDSUB;
"EXIT SUB"    Return EXITSUB;
"DECLARE SUB" Return DECLARESUB;
FUNCTION           Return FUNCTION;
"END FUNCTION"     Return ENDFUNCTION;
"EXIT FUNCTION"    Return EXITFUNCTION;
"DECLARE FUNCTION" Return DECLAREFUNCTION;

	/* For stuff */
FOR        Return FOR;
TO         Return TO;
STEP       Return STEP;
NEXT       Return NEXT;
"EXIT FOR" Return EXITFOR;

	/* Typedef stuff */
INTEGER    Return INTEGER;
SINGLE     Return SINGLE;
DOUBLE     Return DOUBLE;
STRING     Return STRING;
LONG       Return LONG;
AS         Return AS;
TYPE       Return TYPE;
"END TYPE" Return ENDTYPE;
SHARED     Return SHARED;
DIM        Return DIM;
REDIM      Return REDIM;
ERASE      Return ERASE;
CONST      Return CONST;
"OPTION BASE" Return OPTION_BASE;
STATIC     Return STATIC;

	/* Control stuff */
ERROR      Return ERROR;
RESUME     Return RESUME;
GOTO       |
"GO TO"    Return GOTO;
GOSUB      Return GOSUB;
RETURN     Return RETURN;
RESTORE    Return RESTORE;
READ       Return READ;
CALL       Return CALL;
END        Return END;
	
	/* Something */
"DEF SEG"  Return DEFSEG;
BYVAL      Return BYVAL;
CDECL      Return CDECL;
ALIAS      Return ALIAS;
SEG        Return SEG;
COMMON     Return COMMON;

	/* File stuff */
OPEN       Return OPEN;
OUTPUT     Return OUTPUT;
INPUT      Return INPUT;
APPEND     Return APPEND;
RANDOM     Return RANDOM;
BINARY     Return BINARY;
WRITE      Return WRITE;
ACCESS     Return ACCESS;
LOCK       Return LOCK;
CLOSE      Return CLOSE;

	/* Other I/O stuff */
"LINE INPUT"  Return LINEINPUT;
\?\ *\#    |
PRINT\ +\# Return PRINT_F;
LINE\ INPUT\ +\# Return LINEINPUT_F;
INPUT\ +\# Return INPUT_F;
\?         |
PRINT      Return PRINT;
USING      Return USING;
LPRINT     Return LPRINT;

	/* Graphics */
LOCATE     Return LOCATE;
GET        Return GET;
PUT        Return PUT;
SCREEN     Return SCREEN;
PALETTE    Return PALETTE;
"VIEW PRINT"  Return VIEWPRINT;
VIEW       Return VIEW;
PSET       Return PSET;
PRESET     Return PRESET;
LINE       Return LINE;
CIRCLE     Return CIRCLE;
PAINT      Return PAINT;
DRAW       Return DRAW;
COLOR      Return COLOR;
CLS        Return CLS;

	/* Other */
LET        Return LET;
ON         Return ON;
OFF        Return OFF;
LIST       Return LIST;
PLAY       Return PLAY;
STOP       Return STOP;
POKE       Return POKE;
SYSTEM     Return SYSTEM;
OUT        Return OUT;
WAIT       Return WAIT;
"ON KEY"   Return ON_KEY;
"ON PLAY"  Return ON_PLAY;
"ON ERROR" Return ON_ERROR;
"ON TIMER" Return ON_TIMER;

	/* Functions without parameters*/
INKEY\$    Return INKEY_S;
CSRLIN     Return CSRLIN;
FREEFILE   Return FREEFILE;
DATE\$     Return DATE_S;
TIME\$     Return TIME_S;
TIMER      Return TIMER;
ERL        Return ERL;
ERR        Return ERR;
RND        Return RND;
        /* Functions taking strings */
ASC        Return ASC;
LEN        Return LEN;
CVD        Return CVD;
CVI        Return CVI;
CVL        Return CVL;
CVS        Return CVS;
LTRIM\$    Return LTRIM_S;
RTRIM\$    Return RTRIM_S;
LCASE\$    Return LCASE_S;
UCASE\$    Return UCASE_S;
VAL        Return VAL;
        /* Functions with mixed parameters */
ENVIRON\$  Return ENVIRON_S; /* also takes strings */
FILEATTR   Return FILEATTR; /* two args */
FRE        Return FRE;       /* also takes strings */
INSTR      Return INSTR;
MID\$      Return MID_S;
LEFT\$     Return LEFT_S;
RIGHT\$    Return RIGHT_S;
        /* Functions taking numbers */
ABS        Return ABS;
ATN        Return ATN;
CDBL       Return CDBL;
CINT       Return CINT;
CHR\$      Return CHR_S;
CLNG       Return CLNG;
COS        Return COS;
CSNG       Return CSNG;
EXP        Return EXP;
FIX        Return FIX;
HEX\$      Return HEX_S;
INP        Return INP;
INPUT\$    Return INPUT_S;
INT        Return INT;
LOC        Return LOC;
LOF        Return LOF;
LOG        Return LOG;
LPOS       Return LPOS;
MKD\$      Return MKD_S;
MKI\$      Return MKI_S;
MKL\$      Return MKL_S;
MKS\$      Return MKS_S;
OCT\$      Return OCT_S;
POINT      Return POINT; /* gfx */
PEEK       Return PEEK;
POS        Return POS;
SPC        Return SPC;
SQR        Return SQR;
STR\$      Return STR_S;
STRING\$   Return STRING_S;
TAB        Return TAB;
TAN        Return TAN;
SPACE\$    Return SPACE_S;
        /* Meta-functions */
LBOUND     Return LBOUND; /* array */
UBOUND     Return UBOUND; /* array */
VARPTR     Return VARPTR;   /* any */
VARPTR\$   Return VARPTR_S; /* any */
VARSEG     Return VARSEG;   /* any */


DEF        Return DEF;
"END DEF"  Return ENDDEF;
"EXIT DEF" Return EXITDEF;
FN[a-z][a-z0-9]*[!#%$&]? {
    yylval.str = (yytext);
    Return FNIDENTIFIER;
  }

[a-z][a-z0-9]*[!#%$&]? {
    //fprintf(stderr, "Identifier: '%s' (%u)\n", yytext, yyleng);
    yylval.str = (yytext);
    Return IDENTIFIER;
  }

-?{digit}+[%&] {
    //fprintf(stderr, "Integer: '%s'\n", yytext);
    char*endp;
    yylval.number  = (strtoll(yytext,&endp,10));
    yylval.numtype = *endp;
    Return NUMBER;
  }

&H[0-9A-F]+[%&]? {
    //fprintf(stderr, "Hex integer: '%s'\n", yytext);
    char*endp;
    yylval.number  = (strtoll(yytext+2,&endp,16));
    yylval.numtype = *endp ? *endp : '%';
    Return NUMBER;
  }

-?{digit}+(\.{digit}*)?(e[+-]?{digit}+)?[!#]? {
    char*endp;
    yylval.number  = (strtold(yytext,&endp));
    yylval.numtype = *endp ? *endp : '!';
    Return NUMBER;
  }

-?\.{digit}+(e[+-]?{digit}+)?[!#]? {
    char*endp;
    yylval.number  = (strtold(yytext,&endp));
    yylval.numtype = *endp ? *endp : '!';
    Return NUMBER;
  }

#{digit}* {
    ++yytext; --yyleng; // skip '#'
    yylval.number  = (strtoll(yytext,0,10));
    yylval.numtype = '%';
    Return FILENUM;
  }

:[ :]*  BEGIN(INITIAL); Return COLONCHAR;
\n+     BEGIN(INITIAL); Return CRLF;

"<="    Return LESSOREQUAL;
">="    Return GREATEROREQUAL;
"<>"    |
"><"    |
"!="    Return NOTEQUAL;

\"[^"\n]*\"  {
    ++yytext; yyleng -= 2; // skip " and remove "
    //fprintf(stderr, "String: '%.*s'\n", (int)yyleng, yytext);
    yylval.str.assign(yytext, yytext+yyleng);
    Return STRINGCONST;
  }

[-<>=+*/\\,;().^]   Return *yytext;

{blank}+    /* blank   */

<<EOF>> {
    fclose(yyin);
    if(!contexts.empty())
    {
        yyEndFile();
        current_context = contexts.back();
        contexts.pop_back();
        yyin = current_context.fp;
      
        yy_delete_buffer(YY_CURRENT_BUFFER);
        yy_switch_to_buffer(current_context.state);
    }
    else
    {
        yyterminate();
    }
  }

<*>.    BEGIN(INITIAL); Return UNKNOWN_CHARACTER;

%%
